// Emacs will be in -*- Mode: c++ -*-
//
// ************ DO NOT REMOVE THIS BANNER ****************
//
//  Nicolas Di Cesare <Nicolas.Dicesare@free.fr>
//  http://acm.emath.fr/~dicesare
//
//********************************************************
//
//  Optimization Method and Problem abstract classes
//
//********************************************************
#ifndef _optimizer_h
#define _optimizer_h

#include <iostream>
#include <iomanip>

#include <costfunction.h>
#include <criteria.h>

#include <ql/handle.hpp>

//using QuantLib::Handle;
using QuantLib::DotProduct;
using std::setw;

template <class V> class OptimizationMethod;
template <class V> class OptimizationProblem;


/*! 
  Optimization Method abstract class for unconstrained optimization pb
*/
template <class V>
class OptimizationMethod {
public:
  typedef double value_type;
protected:
  //! initial value of unknowns
  V initialValue_;
  //! current iteration step in the Optimization process
  int iterationNumber_;
  //! optimization end criteria
  OptimizationEndCriteria endCriteria_;
  //! number of evaluation of cost function and its gradient
  int functionEvaluation_, gradientEvaluation_;
  //! function and gradient norm values of the last step
  value_type functionValue_, squaredNorm_;
  //! current values of the local minimum and the search direction
  V x_, searchDirection_;
public:
  explicit OptimizationMethod() 
    : iterationNumber_(0),
      functionEvaluation_(0), gradientEvaluation_(0), 
      functionValue_(1), squaredNorm_(1)  {}
  virtual ~OptimizationMethod() {}

  //! Set initial value
  inline void setInitialValue(const V& initialValue) 
  {
    iterationNumber_ = 0;
    initialValue_ = initialValue; 
    x_ = initialValue; 
    searchDirection_ = V(x_.size());
  } 
  //! Set optimization end criteria
  inline void setEndCriteria(const OptimizationEndCriteria& endCriteria) { endCriteria_ = endCriteria;}

  //! initial value of the optimization method
  //inline const V& initialValue() { return initialValue_;}
  //! current iteration number
  inline int& iterationNumber() { return iterationNumber_;}
  //! optimization end criteria
  inline OptimizationEndCriteria& endCriteria() { return endCriteria_;}
  //! number of evaluation of cost function
  inline int & functionEvaluation() { return functionEvaluation_; }
  //! number of evaluation of cost function gradient
  inline int & gradientEvaluation() { return gradientEvaluation_; }
  //! value of cost function
  inline value_type& functionValue() { return functionValue_; }
  //! value of cost function gradient norm
  inline value_type& gradientNormValue() { return squaredNorm_; }
  //! current value of the local minimum
  V& x() { return x_;}
  //! current value of the search direction
  V& searchDirection() { return searchDirection_;}
  
  //! minimize the optimization problem P
  virtual void Minimize(OptimizationProblem<V> & P) = 0;
};



/*!
  Default Optimization problem output. Define your own by simply overloading
  init(..) and save(...)
*/
template <class V>
class OptimizationProblemOutput {
  std::ostream& os_;
public:
  typedef double value_type;

  OptimizationProblemOutput() : os_(std::cout) {}
  OptimizationProblemOutput(std::ostream& os) : os_(os) {}
  virtual ~OptimizationProblemOutput() {}

  virtual void init(OptimizationMethod<V>& method) 
  {
#ifdef DISPLAY_OPTI
    os_ << "Initial value : " << method.x() << std::endl;
    os_ << std::endl
	<< " fen : function evaluation number" << std::endl
	<< " gen : gradient evaluation number" << std::endl
	<< "| iter | function value   |     gradient norm     |  line-search step     |  fen  |  gen  |" << std::endl
	<< "|------|------------------|-----------------------|-----------------------|-------|-------|" << std::endl;
#endif
  }

  virtual void save(int iterationNumber, 
		    value_type function, 
		    value_type normGradient,
		    value_type lineSearchStep,
		    OptimizationMethod<V>& method)
  {
    method.functionValue() = function; method.gradientNormValue() = normGradient;
#ifdef DISPLAY_OPTI
    os_.setf(std::ios::scientific,std::ios::floatfield);
    os_ << "|" << setw(4) << iterationNumber
	<< "  |" << setw(15) << function 
	<< "   |" << setw(19) << normGradient
	<< "    |" << setw(19) << lineSearchStep
	<< "    |" << setw(5) << method.functionEvaluation() 
	<< "  |" << setw(5) << method.gradientEvaluation() 
	<< "  |" << std::endl;
    os_.flush();
#endif
  }
};

/*! 
  Unconstrained optimization pb
*/
template <class V>
class OptimizationProblem {
public:
  typedef double value_type;
protected:
  //! Unconstrained cost function
  CostFunction<V> & costFunction_;
  //! Unconstrained optimization method
  OptimizationMethod<V> & method_;
  QuantLib::Handle<OptimizationProblemOutput<V> > opo_;
public:
  //! default constructor
  OptimizationProblem(CostFunction<V> & f,           // Function and it gradient vector
		      OptimizationMethod<V> & meth)  // Optimization method
    : costFunction_(f), method_(meth), 
      opo_(QuantLib::Handle<OptimizationProblemOutput<V> >(new OptimizationProblemOutput<V>()))
  {
    opo_->init(method_);
  }
  //! default constructor
  OptimizationProblem(CostFunction<V> & f,           // Function and it gradient vector
		      OptimizationMethod<V> & meth,          // Optimization method
		      QuantLib::Handle<OptimizationProblemOutput<V> >& opo)
    : costFunction_(f), method_(meth), opo_(opo)
  {
    opo_->init(method_);
  }

  //! destructor
  ~OptimizationProblem() {}

  //! call cost function computation and increment evaluation counter
  value_type value(const V& x)
  { 
    method_.functionEvaluation()++; 
    return costFunction_.value(x); 
  }

  //! call cost function gradient computation and increment evaluation counter
  void firstDerivative(V& grad_f, const V& x)
  { 
    method_.gradientEvaluation()++; 
    costFunction_.firstDerivative(grad_f, x); 
  }

  //! call cost function computation and it gradient
  value_type valueAndFirstDerivative(V& grad_f, const V& x)
  { 
    method_.functionEvaluation()++; method_.gradientEvaluation()++; 
    return costFunction_.valueAndFirstDerivative(grad_f, x);
  }

  //! Unconstrained optimization method
  OptimizationMethod<V> & optimisationMethod() { return method_;}

  //! Minimization
  void Minimize()
  { 
    method_.Minimize( *this ); 
  }

  V& minimumValue() { return method_.x();}

  //! Save results
  void Save(int iterationNumber, 
	    value_type function, 
	    value_type normGradient,
	    value_type lineSearchStep,
	    OptimizationMethod<V>& method)
  {
    opo_->save(iterationNumber, function, normGradient, lineSearchStep, method);
  }
};

#endif
